Namespace Gauss

    Public Class Ex
        Inherits Exception
        Public Sub New(ByVal msg As String)
            MsgBox("    : " & vbNewLine + msg)
        End Sub
    End Class

    Public Class LinearSystem
        Private initial_a_matrix(,) As Double
        Private a_matrix(,) As Double   '  A
        Private x_vector() As Double   '   x
        Private initial_b_vector() As Double   ' 
        Private b_vector() As Double   '   b
        Private u_vector() As Double   '    U  
        Private eps As Double = 0.0001         '       
        Private size As Integer             '  

        Public Sub LinearSystem(ByVal a_matrix(,) As Double, ByVal b_vector() As Double)

            Dim b_length As Integer = b_vector.Length
            Dim a_length As Integer = a_matrix.Length
            If (a_length <> b_length * b_length) Then _
                Throw New ArgumentException("      A        B.")

            Me.initial_a_matrix = a_matrix '   
            Me.a_matrix = a_matrix.Clone() '      
            Me.initial_b_vector = b_vector '  
            Me.b_vector = b_vector.Clone() '     
            ReDim Me.x_vector(b_length)
            ReDim Me.u_vector(b_length)
            Me.size = b_length
            Me.eps = eps

            GaussSolve()
        End Sub

        Public ReadOnly Property XVector() As Double()
            Get
                Return x_vector
            End Get
        End Property

        Public ReadOnly Property UVector() As Double()
            Get
                Return u_vector
            End Get
        End Property

        '    
        Private Function InitIndex() As Integer()
            Dim index() As Integer, i As Integer
            ReDim index(size)
            For i = 0 To index.Length - 1
                index(i) = i
            Next
            Return index
        End Function

        '     
        Private Function FindR(ByVal row As Integer, ByVal index() As Integer) As Double
            Dim max_index As Integer = row
            Dim max As Double = a_matrix(row, index(max_index))
            Dim max_abs As Double = Math.Abs(max)
            Dim cur_index As Integer
            Dim cur, cur_abs As Double

            For cur_index = row + 1 To size - 1
                cur = a_matrix(row, index(cur_index))
                cur_abs = Math.Abs(cur)
                If (cur_abs > max_abs) Then
                    max_index = cur_index
                    max = cur
                    max_abs = cur_abs
                End If
            Next

            If (max_abs < eps) Then
                If (Math.Abs(b_vector(row)) > eps) Then
                    Throw New Ex("  .")
                Else
                    Throw New Ex("    .")
                End If
            End If

            '    
            Dim temp As Integer = index(row)
            index(row) = index(max_index)
            index(max_index) = temp

            Return max
        End Function

        '     
        Private Sub GaussSolve()
            Dim index() As Integer = InitIndex()
            GaussForwardStroke(index)
            GaussBackwardStroke(index)
            GaussDiscrepancy()
        End Sub

        '    
        Private Sub GaussForwardStroke(ByVal index() As Integer)
            Dim i, k, j As Integer, r, p As Double
            '     
            For i = 0 To size - 1
                ' 1)   
                r = FindR(i, index)

                ' 2)     A
                For j = 0 To size - 1
                    a_matrix(i, j) /= r
                Next

                ' 3)  i-   b
                b_vector(i) /= r

                ' 4)       
                For k = i + 1 To size - 1
                    p = a_matrix(k, index(i))
                    For j = i To size - 1
                        a_matrix(k, index(j)) -= a_matrix(i, index(j)) * p
                    Next
                    b_vector(k) -= b_vector(i) * p
                    a_matrix(k, index(i)) = 0.0
                Next
            Next
        End Sub

        '    
        Private Sub GaussBackwardStroke(ByVal index() As Integer)
            Dim i, j As Integer
            Dim x_i As Double
            '      
            For i = size - 1 To i Step -1
                ' 1)     x
                x_i = b_vector(i)
                ' 2)   
                For j = i + 1 To size - 1
                    x_i -= x_vector(index(j)) * a_matrix(i, index(j))
                Next
                x_vector(index(i)) = x_i
            Next

        End Sub
 
        '  
        ' U = b - x * A
        ' x -  ,   
        Private Sub GaussDiscrepancy()
            Dim i, j As Integer
            Dim actual_b_i As Double
            For i = 0 To size - 1
                actual_b_i = 0.0    '   i- 
                '     x
                For j = 0 To size - 1
                    actual_b_i += initial_a_matrix(i, j) * x_vector(j)
                Next
                ' i-   
                u_vector(i) = initial_b_vector(i) - actual_b_i
            Next
        End Sub

    End Class


End Namespace
